@@ -43,6 +43,17 @@ func TestTargetLabels(t *testing.T) {
 	want := labels.FromStrings(model.JobLabel, "some_job", "foo", "bar")
 	got := target.Labels()
 	require.Equal(t, want, got)
+	i := 0
+	target.LabelsRange(func(l labels.Label) {
+		switch i {
+		case 0:
+			require.Equal(t, labels.Label{Name: "foo", Value: "bar"}, l)
+		case 1:
+			require.Equal(t, labels.Label{Name: model.JobLabel, Value: "some_job"}, l)
+		}
+		i++
+	})
+	require.Equal(t, 2, i)
 }
 
 func TestTargetOffset(t *testing.T) {
@@ -123,13 +134,13 @@ func TestTargetURL(t *testing.T) {
 	require.Equal(t, expectedURL, target.URL())
 }
 
-func newTestTarget(targetURL string, deadline time.Duration, lbls labels.Labels) *Target {
+func newTestTarget(targetURL string, _ time.Duration, lbls labels.Labels) *Target {
 	lb := labels.NewBuilder(lbls)
 	lb.Set(model.SchemeLabel, "http")
 	lb.Set(model.AddressLabel, strings.TrimPrefix(targetURL, "http://"))
 	lb.Set(model.MetricsPathLabel, "/metrics")
 
-	return &Target{labels: lb.Labels(labels.EmptyLabels())}
+	return &Target{labels: lb.Labels()}
 }
 
 func TestNewHTTPBearerToken(t *testing.T) {
